# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .backend import ArrayLike
from .dimension import Dimension
from .trainshape import TrainShape, change_dims
from .trainbase import TrainBase

def fuse[T: ArrayLike](*trains: TrainBase[T]) -> TrainBase[T]:
    """Concatenates trains forming their outer product"""
    if len(trains) < 2:
        raise ValueError("fuse() requires at least two trains")

    all_dims = []
    all_digits = []
    all_data = []

    for train in trains:
        # copy shape and dimensions
        dims = [Dimension([d.base for d in dim]) for dim in train.shape.dims]
        shape = change_dims(train.shape, dims)

        all_dims.extend(dims)
        all_digits.extend(shape.digits)
        all_data.extend(train.data)

    shape = TrainShape(all_dims, all_digits)
    return TrainBase(shape, all_data)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence

from .backend import ArrayLike, ArrayNamespace
from .dimension import Dimension
from .matrixdecomposition import MatrixDecomposition
from .trainbase import TrainBase
from .trainshape import TrainShape

from .decompositionaddition import DecompositionAddition
from .variationaladdition import VariationalAddition
from .svdecomposition import SVDecomposition, SVDecompositionResult
from .sweepingstrategy import SweepingStrategy
from .tensordecomposition import TensorDecomposition
from .trainbase import QRDecomposition

from .einsumexpression import EinsumExpression
from .full import full
from .matmul import matmul
from .toeplitz import shift


def wavelet[T: ArrayLike](
    xp: ArrayNamespace[T],
    dim: Dimension,
    coeffs: Sequence[float],
    decomp: MatrixDecomposition = SVDecomposition(max_rank=8),
) -> TrainBase:
    n_input = dim.size()
    n_coeffs = len(coeffs)
    assert n_input >= n_coeffs

    strat = SweepingStrategy(ncores=2, nsweeps=1)

    permutated_shifts = [permutated_shift(xp, dim, i) for i in range(n_coeffs)]

    coefficients = [
        get_ith_coffencients_tensor(xp, dim, coeffs, i, decomp) for i in range(n_coeffs)
    ]

    expr = EinsumExpression[T](
        "ab,a->ab", permutated_shifts[0].shape, coefficients[0].shape, method="exact"
    )

    X = tuple(expr(P, C) for P, C in zip(permutated_shifts, coefficients))

    _add = DecompositionAddition(strat, decomp)
    guess = _add(*X)  # type: ignore

    var_add = VariationalAddition(guess.shape, decomp, strat)
    guess = var_add(guess, *X)  # type: ignore
    return guess


def permutated_shift[T: ArrayLike](
    xp: ArrayNamespace[T], dim: Dimension, shift_by: int
) -> TrainBase:
    dim_2 = Dimension([*[d.base for d in dim[1:]], dim[0].base])

    shape = TrainShape([dim, dim_2], [digits for digits in zip(dim, dim_2)])
    decomp = TensorDecomposition(QRDecomposition[T]())

    P = shift(xp, dim, shift_by // 2, circular=True)
    data = []
    tmp = P.data[0]

    if shift_by % 2 == 1:
        tmp = xp.flip(tmp, axis=2)

    for i in range(len(P.data) - 1):
        supercore = xp.tensordot(tmp, P.data[i + 1], axes=([3], [0]))
        supercore = xp.permute_dims(supercore, (0, 1, 4, 3, 2, 5))
        res = decomp.left(supercore, 3)
        data.append(res.left)
        tmp = res.right
    data.append(tmp)

    return TrainBase(shape, data)


def coefficients_tensor[T: ArrayLike](
    xp: ArrayNamespace[T],
    dim: Dimension,
    c1: float,
    c2: float,
    c3: float,
    c4: float,
    shift_by: int,
    decomp: MatrixDecomposition,
) -> TrainBase:
    strat = SweepingStrategy(ncores=2, nsweeps=1)

    shape = TrainShape(dim, [(d,) for d in dim])  # [[*dim]] = [[d1, d2, ...]]

    n_input = dim.size()
    n_half = n_input // 2

    # shift forward and backward to cut and put in correct position
    # [ 0  |  n_half - shift_by  |  h_half  |  n_input - shift_by]
    # T1 = [c1 | 0 | 0 | 0], T3 = [0 | 0 | c3 | 0]
    # T2 = [0 | c2 | 0 | 0], T4 = [0 | 0 | 0 | c4]

    T1 = matmul(
        full(xp, shape, c1),
        matmul(shift(xp, dim, n_half + shift_by), shift(xp, dim, -n_half - shift_by)),
    )
    T3 = matmul(
        full(xp, shape, c3),
        matmul(shift(xp, dim, n_half + shift_by), shift(xp, dim, -shift_by)),
    )

    Ts = [T1, T3]

    if shift_by > 0:
        T2 = matmul(
            matmul(full(xp, shape, c2), shift(xp, dim, n_input - shift_by)),
            shift(xp, dim, -n_half),
        )
        T4 = matmul(full(xp, shape, c4), shift(xp, dim, n_input - shift_by))

        Ts.append(T2)
        Ts.append(T4)

    _add = DecompositionAddition[T, SVDecompositionResult](strat, decomp)
    guess = _add(*Ts)
    # var_add = VariationalAddition(guess.shape, decomp, strat)
    # guess = var_add(guess, *Ts)
    return guess


def get_ith_coffencients_tensor[T: ArrayLike](
    xp: ArrayNamespace[T],
    dim: Dimension,
    coeffs: Sequence[float],
    i: int,
    decomp: MatrixDecomposition,
) -> TrainBase:
    n_coeffs = len(coeffs)

    is_odd = i % 2 == 1
    odd = 1 - 2 * is_odd  # multiply by -1 if odd

    ix = i + 2 * (not (is_odd))
    iy = ix - 1
    # ix: 0 -> 2, 1 -> 1, 2 -> 4, 3 -> 3, 4 -> 6, 5 -> 5, ...
    # iy: 0 -> 1, 1 -> 0, 2 -> 3, 3 -> 2, 4 -> 5, 5 -> 4, ...

    c1 = coeffs[i]
    c2 = coeffs[iy]
    c3 = coeffs[-ix] * -odd
    c4 = coeffs[n_coeffs - i - 1] * odd

    return coefficients_tensor(xp, dim, c1, c2, c3, c4, i // 2, decomp)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from math import prod, ceil

from .backend import ArrayNamespace, get_index_dtype, ArrayLike
from .dimension import Dimension
from .trainbase import TrainBase
from .trainshape import TrainShape
from .exactaddition import ExactAddition

def slice_operator[T: ArrayLike](xp: ArrayNamespace[T], dim: Dimension, step: int) -> TrainBase[T]:
    tmp = step
    idx = 0
    for i in range(len(dim)-1, -1, -1):
        tmp //= dim[i].base
        if tmp == 1:
            idx = i
            break

    if dim.size() / step < step:
        return sum_slice(xp, dim, step)
    elif prod(digit.base for digit in dim[idx:]) % step != 0:
        return factor_slice(xp, dim, step)
    else:
        return factor_slice_exact(xp, dim, step)

def sum_slice(xp: ArrayNamespace, dim: Dimension, step: int):
    idxs = xp.asarray([step], dtype=get_index_dtype(xp))[:,xp.newaxis]
    digits = dim.to_digits(idxs)[:,0]

    add = ExactAddition()

    train = val_cores(xp, dim, 0, 1.0)
    for i in range(1, ceil(dim.size()/step)):
        train = add(train, val_cores(xp, dim, i*step, 1.0))

    return train

def val_cores[T: ArrayLike](xp: ArrayNamespace[T], dim: Dimension, idx: int, val: float) -> TrainBase[T]:
    idxs = xp.asarray([idx], dtype=get_index_dtype(xp))[:,xp.newaxis]
    digits = dim.to_digits(idxs)[:,0]
    cores = []
    for i in range(len(dim)):
        core = xp.zeros((1, dim[i].base, 1))
        core[0,digits[i],0] = 1.0
        cores.append(core)
    cores[0] *= val
    shape = TrainShape(dim, [(d,) for d in dim])
    return TrainBase(shape, cores, copy_data=False)


def factor_slice[T: ArrayLike](xp: ArrayNamespace[T], dim: Dimension, step: int) -> TrainBase[T]:
    cores = []
    offs = xp.asarray([0])

    tmp = dim.size()
    for digit in dim:
        tmp //= digit.base

        noffs = len(offs)

        # generate indices
        left = xp.arange(noffs)[:,None]
        mid = xp.arange(digit.base)[None,:]
        right = (offs[:,None] - tmp*mid) % step

        # get unique columns indices
        roffs = xp.unique_all(right)[0]
        nright = xp.searchsorted(roffs, right)

        # generate matrix
        mat = xp.zeros((noffs, digit.base, roffs.size))
        mat[left, mid, nright] = 1.0
        cores.append(mat)
        offs = roffs
    cores[-1] = cores[-1][:,:,0][:,:,None]

    # generate vector
    shape = TrainShape(dim, [(d,) for d in dim])
    return TrainBase(shape, cores, copy_data=False)

def factor_slice_exact[T: ArrayLike](xp: ArrayNamespace[T], dim: Dimension, step: int) -> TrainBase[T]:
    tmp = step
    idx = 0
    for i in range(len(dim)-1, -1, -1):
        tmp //= dim[i].base
        if tmp == 1:
            idx = i
            break

    if prod(digit.base for digit in dim[idx:]) % step != 0:
        raise ValueError("period must be divisible by dimension size")

    idxs = xp.asarray([step], dtype=get_index_dtype(xp))[:,xp.newaxis]
    digits = dim.to_digits(idxs)[:,0]

    cores = []
    for i in range(len(dim)):
        core = xp.zeros((1, dim[i].base, 1))
        if i < idx:
            core[0,:,0] = 1.0
        else:
            core[0,digits[i],0] = 1.0
        cores.append(core)

    shape = TrainShape(dim, [(d,) for d in dim])
    return TrainBase(shape, cores, copy_data=False)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from math import exp as mexp

from .backend import ArrayNamespace, ArrayLike
from .uniformgrid import UniformGrid
from .trainbase import TrainBase
from .trainshape import TrainShape

# for further reading please consider https://arxiv.org/abs/2602.20226 section 4.1 "Exponential functions"
# see equation 21

def exp[T: ArrayLike](xp: ArrayNamespace[T], grid: UniformGrid, factor: float, offset: float) -> TrainBase[T]:
    if len(grid.dims) != 1:
        raise ValueError("Exponential function only supports 1D uniform grid")
    dim = grid.dims[0]
    shape = TrainShape(dim, [(d,) for d in dim])
    domain = grid.domains[0]
    spacing = grid.spacings[0]
    
    data = []
    for digit in dim:
        xi = spacing*digit.factor
        tmp = xp.zeros((1, digit.base, 1))
        for j in range(digit.base):
            tmp[:,j,:] = mexp(factor * j*xi)
        data.append(tmp)
    data[0] *= mexp(factor * (domain.lower - offset))
    return TrainBase(shape, data, copy_data=False)



import unittest
from itertools import product

from trainsum import TrainSum
from utils import backends

from typing import Sequence
import pywt


class TestWavelet(unittest.TestCase):
    def setUp(self):
        self.coeffs = [pywt.Wavelet(db).dec_lo[::-1] for db in ["db2", "db4", "db20"]]
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [64, 100, 128, 2**10]

    def ground_truth(self, xp, coefficients: Sequence[float], n: int):
        n_coeffs = len(coefficients)
        mat = xp.zeros((n, n))
        for i, j in enumerate(range(0, n, 2)):
            js = [x % n for x in range(j, j + n_coeffs)]
            mat[i, js] = coefficients
        odds = xp.ones(n_coeffs)
        odds[1::2] = odds[1::2] * -1
        rcoeffs = coefficients[::-1] * odds
        for i, j in enumerate(range(0, n, 2)):
            i += n // 2
            js = [x % n for x in range(j, j + n_coeffs)]
            mat[i, js] = rcoeffs
        return mat

    def test_wavelet(self):

        for coeffs, ts, size in product(self.coeffs, self.trainsum, self.sizes):
            decomp = ts.svdecomposition(max_rank=18, cutoff=1e-10)
            xp = ts.namespace
            dim = ts.dimension(size)

            train = ts.wavelet(dim, coeffs, decomp)
            approx = train.to_tensor()

            exact = self.ground_truth(xp, coeffs, size)
            diff = abs(xp.sum((exact - approx) ** 2))
            self.assertLess(diff, 1e-10)


if __name__ == "__main__":
    unittest.main()
